@@ -1,6 +1,7 @@
 /* GObject introspection: Typelib creation
  *
  * Copyright (C) 2005 Matthias Clasen
+ * Copyright (C) 2008,2009 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -184,6 +185,9 @@ g_ir_node_new (GIrNodeTypeId type)
     }
 
   node->type = type;
+  node->offset = 0;
+  node->attributes = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                            g_free, g_free);
 
   return node;
 }
@@ -261,6 +265,7 @@ g_ir_node_free (GIrNode *node)
 	GIrNodeVFunc *vfunc = (GIrNodeVFunc *)node;
 	
 	g_free (node->name);
+	g_free (vfunc->invoker);
 	for (l = vfunc->parameters; l; l = l->next)
 	  g_ir_node_free ((GIrNode *)l->data);
 	g_list_free (vfunc->parameters);
@@ -285,7 +290,9 @@ g_ir_node_free (GIrNode *node)
 	g_free (node->name);
 	g_free (iface->gtype_name);
 	g_free (iface->gtype_init);
-	
+
+
+	g_free (iface->glib_type_struct);
 	g_free (iface->parent);
 
 	for (l = iface->interfaces; l; l = l->next)
@@ -398,6 +405,8 @@ g_ir_node_free (GIrNode *node)
       break;
     } 
 
+  g_hash_table_destroy (node->attributes);
+
   g_free (node);
 }
 
@@ -411,7 +420,7 @@ g_ir_node_get_size (GIrNode *node)
   switch (node->type)
     {
     case G_IR_NODE_CALLBACK:
-      size = 12; 
+      size = sizeof (CallbackBlob);
       break;
 
     case G_IR_NODE_FUNCTION:
@@ -419,11 +428,12 @@ g_ir_node_get_size (GIrNode *node)
       break;
 
     case G_IR_NODE_PARAM:
-      size = 12;
+      /* See the comment in the G_IR_NODE_PARAM/ArgBlob writing below */
+      size = sizeof (ArgBlob) - sizeof (SimpleTypeBlob);
       break;
 
     case G_IR_NODE_TYPE:
-      size = 4;
+      size = sizeof (SimpleTypeBlob);
       break;
 
     case G_IR_NODE_OBJECT:
@@ -431,7 +441,7 @@ g_ir_node_get_size (GIrNode *node)
 	GIrNodeInterface *iface = (GIrNodeInterface *)node;
 
 	n = g_list_length (iface->interfaces);
-	size = 32 + 2 * (n + (n % 2));
+	size = sizeof (ObjectBlob) + 2 * (n + (n % 2));
 
 	for (l = iface->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
@@ -443,7 +453,7 @@ g_ir_node_get_size (GIrNode *node)
 	GIrNodeInterface *iface = (GIrNodeInterface *)node;
 
 	n = g_list_length (iface->prerequisites);
-	size = 28 + 2 * (n + (n % 2));
+	size = sizeof (InterfaceBlob) + 2 * (n + (n % 2));
 
 	for (l = iface->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
@@ -455,21 +465,21 @@ g_ir_node_get_size (GIrNode *node)
       {
 	GIrNodeEnum *enum_ = (GIrNodeEnum *)node;
 	
-	size = 20;
+	size = sizeof (EnumBlob);
 	for (l = enum_->values; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
       }
       break;
 
     case G_IR_NODE_VALUE:
-      size = 12;
+      size = sizeof (ValueBlob);
       break;
 
     case G_IR_NODE_STRUCT:
       {
 	GIrNodeStruct *struct_ = (GIrNodeStruct *)node;
 
-	size = 24;
+	size = sizeof (StructBlob);
 	for (l = struct_->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
       }
@@ -479,34 +489,34 @@ g_ir_node_get_size (GIrNode *node)
       {
 	GIrNodeBoxed *boxed = (GIrNodeBoxed *)node;
 
-	size = 24;
+	size = sizeof (StructBlob);
 	for (l = boxed->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
       }
       break;
 
     case G_IR_NODE_PROPERTY:
-      size = 12;
+      size = sizeof (PropertyBlob);
       break;
 
     case G_IR_NODE_SIGNAL:
-      size = 12;
+      size = sizeof (SignalBlob);
       break;
 
     case G_IR_NODE_VFUNC:
-      size = 16;
+      size = sizeof (VFuncBlob);
       break;
 
     case G_IR_NODE_FIELD:
-      size = 12;
+      size = sizeof (FieldBlob);
       break;
 
     case G_IR_NODE_CONSTANT:
-      size = 20;
+      size = sizeof (ConstantBlob);
       break;
 
     case G_IR_NODE_ERROR_DOMAIN:
-      size = 16;
+      size = sizeof (ErrorDomainBlob);
       break;
 
     case G_IR_NODE_XREF:
@@ -517,7 +527,7 @@ g_ir_node_get_size (GIrNode *node)
       {
 	GIrNodeUnion *union_ = (GIrNodeUnion *)node;
 
-	size = 32;
+	size = sizeof (UnionBlob);
 	for (l = union_->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
 	for (l = union_->discriminators; l; l = l->next)
@@ -537,6 +547,18 @@ g_ir_node_get_size (GIrNode *node)
   return size;
 }
 
+static void
+add_attribute_size (gpointer key, gpointer value, gpointer data)
+{
+  const gchar *key_str = key;
+  const gchar *value_str = value;
+  gint *size_p = data;
+
+  *size_p += sizeof (AttributeBlob);
+  *size_p += ALIGN_VALUE (strlen (key_str) + 1, 4);
+  *size_p += ALIGN_VALUE (strlen (value_str) + 1, 4);
+}
+
 /* returns the full size of the blob including variable-size parts */
 static guint32
 g_ir_node_get_full_size_internal (GIrNode *parent,
@@ -556,7 +578,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
     case G_IR_NODE_CALLBACK:
       {
 	GIrNodeFunction *function = (GIrNodeFunction *)node;
-	size = 12; 
+	size = sizeof (CallbackBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	for (l = function->parameters; l; l = l->next)
 	  {
@@ -569,7 +591,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
     case G_IR_NODE_FUNCTION:
       {
 	GIrNodeFunction *function = (GIrNodeFunction *)node;
-	size = 24;
+	size = sizeof (FunctionBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	size += ALIGN_VALUE (strlen (function->symbol) + 1, 4);
 	for (l = function->parameters; l; l = l->next)
@@ -582,7 +604,8 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeParam *param = (GIrNodeParam *)node;
 	
-	size = 16;
+	/* See the comment in the G_IR_NODE_PARAM/ArgBlob writing below */
+	size = sizeof (ArgBlob) - sizeof (SimpleTypeBlob);
 	if (node->name)
 	  size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	size += g_ir_node_get_full_size_internal (node, (GIrNode *)param->type);	
@@ -592,9 +615,8 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
     case G_IR_NODE_TYPE:
       {
 	GIrNodeType *type = (GIrNodeType *)node;
-	if (type->tag < GI_TYPE_TAG_ARRAY) 
-	  size = 4;
-	else
+        size = sizeof (SimpleTypeBlob);
+        if (type->tag >= GI_TYPE_TAG_ARRAY)
 	  {
 	    g_debug ("node %p type tag '%s'", node,
 		     g_type_tag_to_string (type->tag));
@@ -602,21 +624,21 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
 	    switch (type->tag)
 	      {
 	      case GI_TYPE_TAG_ARRAY:
-		size = 4 + 4;
+		size = sizeof (ArrayTypeBlob);
 		if (type->parameter_type1)
 		  size += g_ir_node_get_full_size_internal (node, (GIrNode *)type->parameter_type1);
 		break;
 	      case GI_TYPE_TAG_INTERFACE:
-		size = 4 + 4;
+		size += sizeof (InterfaceTypeBlob);
 		break;
 	      case GI_TYPE_TAG_GLIST:
 	      case GI_TYPE_TAG_GSLIST:
-		size = 4 + 4;
+		size += sizeof (ParamTypeBlob);
 		if (type->parameter_type1)
 		  size += g_ir_node_get_full_size_internal (node, (GIrNode *)type->parameter_type1);
 		break;
 	      case GI_TYPE_TAG_GHASH:
-		size = 4 + 4 + 4;
+		size += sizeof (ParamTypeBlob) * 2;
 		if (type->parameter_type1)
 		  size += g_ir_node_get_full_size_internal (node, (GIrNode *)type->parameter_type1);
 		if (type->parameter_type2)
@@ -631,7 +653,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
 		  else
 		    n = 0;
 
-		  size = 4 + 4 + 2 * (n + n % 2);
+		  size += sizeof (ErrorTypeBlob) + 2 * (n + n % 2);
 		}
 		break;
 	      default:
@@ -647,14 +669,15 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
 	GIrNodeInterface *iface = (GIrNodeInterface *)node;
 
 	n = g_list_length (iface->interfaces);
-	size = 32;
+	size = sizeof(ObjectBlob);
 	if (iface->parent)
 	  size += ALIGN_VALUE (strlen (iface->parent) + 1, 4);
+        if (iface->glib_type_struct)
+          size += ALIGN_VALUE (strlen (iface->glib_type_struct) + 1, 4);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	size += ALIGN_VALUE (strlen (iface->gtype_name) + 1, 4);
 	if (iface->gtype_init)
 	  size += ALIGN_VALUE (strlen (iface->gtype_init) + 1, 4);
-	size += ALIGN_VALUE ( + 1, 4);
 	size += 2 * (n + (n % 2));
 
 	for (l = iface->members; l; l = l->next)
@@ -667,7 +690,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
 	GIrNodeInterface *iface = (GIrNodeInterface *)node;
 
 	n = g_list_length (iface->prerequisites);
-	size = 28;
+	size = sizeof (InterfaceBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	size += ALIGN_VALUE (strlen (iface->gtype_name) + 1, 4);
 	size += ALIGN_VALUE (strlen (iface->gtype_init) + 1, 4);
@@ -683,7 +706,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeEnum *enum_ = (GIrNodeEnum *)node;
 	
-	size = 20;
+	size = sizeof (EnumBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	if (enum_->gtype_name)
 	  {
@@ -698,7 +721,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
 
     case G_IR_NODE_VALUE:
       {
-	size = 12;
+	size = sizeof (ValueBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
       }
       break;
@@ -707,7 +730,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeStruct *struct_ = (GIrNodeStruct *)node;
 
-	size = 24;
+	size = sizeof (StructBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	if (struct_->gtype_name)
 	  size += ALIGN_VALUE (strlen (struct_->gtype_name) + 1, 4);
@@ -722,7 +745,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeBoxed *boxed = (GIrNodeBoxed *)node;
 
-	size = 24;
+	size = sizeof (StructBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	if (boxed->gtype_name)
 	  {
@@ -738,7 +761,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeProperty *prop = (GIrNodeProperty *)node;
 	
-	size = 12;
+	size = sizeof (PropertyBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	size += g_ir_node_get_full_size_internal (node, (GIrNode *)prop->type);	
       }
@@ -748,7 +771,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeSignal *signal = (GIrNodeSignal *)node;
 
-	size = 12;
+	size = sizeof (SignalBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	for (l = signal->parameters; l; l = l->next)
 	  size += g_ir_node_get_full_size_internal (node, (GIrNode *)l->data);
@@ -760,7 +783,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeVFunc *vfunc = (GIrNodeVFunc *)node;
 
-	size = 16;
+	size = sizeof (VFuncBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	for (l = vfunc->parameters; l; l = l->next)
 	  size += g_ir_node_get_full_size_internal (node, (GIrNode *)l->data);
@@ -772,7 +795,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeField *field = (GIrNodeField *)node;
 
-	size = 12;
+	size = sizeof (FieldBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	size += g_ir_node_get_full_size_internal (node, (GIrNode *)field->type);	
       }
@@ -782,7 +805,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeConstant *constant = (GIrNodeConstant *)node;
 
-	size = 20;
+	size = sizeof (ConstantBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	/* FIXME non-string values */
 	size += ALIGN_VALUE (strlen (constant->value) + 1, 4);
@@ -794,7 +817,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeErrorDomain *domain = (GIrNodeErrorDomain *)node;
 
-	size = 16;
+	size = sizeof (ErrorDomainBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	size += ALIGN_VALUE (strlen (domain->getquark) + 1, 4);
       }
@@ -814,7 +837,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeUnion *union_ = (GIrNodeUnion *)node;
 
-	size = 32;
+	size = sizeof (UnionBlob);
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	if (union_->gtype_name)
 	  size += ALIGN_VALUE (strlen (union_->gtype_name) + 1, 4);
@@ -847,6 +870,14 @@ g_ir_node_get_full_size (GIrNode *node)
   return g_ir_node_get_full_size_internal (NULL, node);
 }
 
+guint32
+g_ir_node_get_attribute_size (GIrNode *node)
+{
+  guint32 size = 0;
+  g_hash_table_foreach (node->attributes, add_attribute_size, &size);
+  return size;
+}
+
 int
 g_ir_node_cmp (GIrNode *node,
 		GIrNode *other)
@@ -1156,6 +1187,30 @@ g_ir_find_node (GIrModule  *module,
   return node != NULL;
 }
 
+static int
+get_index_of_member_type (GIrNodeInterface *node,
+                          GIrNodeTypeId type,
+                          const char *name)
+{
+  guint index = -1;
+  GList *l;
+
+  for (l = node->members; l; l = l->next)
+    {
+      GIrNode *node = l->data;
+
+      if (node->type != type)
+        continue;
+
+      index++;
+
+      if (strcmp (node->name, name) == 0)
+        break;
+    }
+
+  return index;
+}
+
 static void
 serialize_type (GIrModule    *module, 
 		GList        *modules,
@@ -1284,11 +1339,8 @@ static void
 g_ir_node_build_members (GList         **members,
 			 GIrNodeTypeId   type,
 			 guint16        *count,
-			 GIrModule      *module,
-			 GList          *modules,
-			 GHashTable     *strings,
-			 GHashTable     *types,
-			 guchar         *data,
+			 GIrNode        *parent,
+                         GIrTypelibBuild *build,
 			 guint32        *offset,
 			 guint32        *offset2)
 {
@@ -1302,8 +1354,7 @@ g_ir_node_build_members (GList         **members,
       if (member->type == type)
 	{
 	  (*count)++;
-	  g_ir_node_build_typelib (member, module, modules, strings,
-				   types, data, offset, offset2);
+	  g_ir_node_build_typelib (member, parent, build, offset, offset2);
 	  *members = g_list_delete_link (*members, l);
 	}
       l = next;
@@ -1340,15 +1391,17 @@ g_ir_node_check_unhandled_members (GList         **members,
 }
 
 void
-g_ir_node_build_typelib (GIrNode    *node,
-			 GIrModule  *module,
-			 GList      *modules,
-			 GHashTable *strings,
-			 GHashTable *types,
-			 guchar     *data,
-			 guint32    *offset,
-			 guint32    *offset2)
+g_ir_node_build_typelib (GIrNode         *node,
+                         GIrNode         *parent,
+                         GIrTypelibBuild *build,
+                         guint32         *offset,
+                         guint32         *offset2)
 {
+  GIrModule *module = build->module;
+  GList *modules = build->modules;
+  GHashTable *strings = build->strings;
+  GHashTable *types = build->types;
+  guchar *data = build->data;
   GList *l;
   guint32 old_offset = *offset;
   guint32 old_offset2 = *offset2;
@@ -1362,6 +1415,15 @@ g_ir_node_build_typelib (GIrNode    *node,
 
   g_ir_node_compute_offsets (node, module, modules);
 
+  /* We should only be building each node once.  If we do a typelib expansion, we also
+   * reset the offset in girmodule.c.
+   */
+  g_assert (node->offset == 0);
+  node->offset = *offset;
+  build->offset_ordered_nodes = g_list_prepend (build->offset_ordered_nodes, node);
+
+  build->n_attributes += g_hash_table_size (node->attributes);
+
   switch (node->type)
     {
     case G_IR_NODE_TYPE:
@@ -1369,7 +1431,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	GIrNodeType *type = (GIrNodeType *)node;
 	SimpleTypeBlob *blob = (SimpleTypeBlob *)&data[*offset];
 
-	*offset += 4;
+	*offset += sizeof (SimpleTypeBlob);
 	
 	if (type->tag < GI_TYPE_TAG_ARRAY ||
 	    type->tag == GI_TYPE_TAG_UTF8 ||
@@ -1425,19 +1487,18 @@ g_ir_node_build_typelib (GIrNode    *node,
                       else
                         array->length = -1;
 		      
-		      pos = *offset2 + 4;
-		      *offset2 += 8;
+		      pos = *offset2 + G_STRUCT_OFFSET (ArrayTypeBlob, type);
+		      *offset2 += sizeof (ArrayTypeBlob);
 		      
-		      g_ir_node_build_typelib ((GIrNode *)type->parameter_type1, 
-					       module, modules, strings, types, 
-					       data, &pos, offset2);
+		      g_ir_node_build_typelib ((GIrNode *)type->parameter_type1,
+		                               node, build, &pos, offset2);
 		    }
 		    break;
 		    
 		  case GI_TYPE_TAG_INTERFACE:
 		    {
 		      InterfaceTypeBlob *iface = (InterfaceTypeBlob *)&data[*offset2];
-		      *offset2 += 4;
+		      *offset2 += sizeof (InterfaceTypeBlob);
 
 		      iface->pointer = type->is_pointer;
 		      iface->reserved = 0;
@@ -1460,12 +1521,11 @@ g_ir_node_build_typelib (GIrNode    *node,
 		      param->reserved2 = 0;
 		      param->n_types = 1;
 		      
-		      pos = *offset2 + 4;
-		      *offset2 += 8;
+		      pos = *offset2 + G_STRUCT_OFFSET (ParamTypeBlob, type);
+		      *offset2 += sizeof (ParamTypeBlob) + sizeof (SimpleTypeBlob);
 		      
 		      g_ir_node_build_typelib ((GIrNode *)type->parameter_type1, 
-					       module, modules, strings, types,
-					       data, &pos, offset2);
+					       node, build, &pos, offset2);
 		    }
 		    break;
 		    
@@ -1480,15 +1540,13 @@ g_ir_node_build_typelib (GIrNode    *node,
 		      param->reserved2 = 0;
 		      param->n_types = 2;
 		      
-		      pos = *offset2 + 4;
-		      *offset2 += 12;
+		      pos = *offset2 + G_STRUCT_OFFSET (ParamTypeBlob, type);
+		      *offset2 += sizeof (ParamTypeBlob) + sizeof (SimpleTypeBlob)*2;
 		      
 		      g_ir_node_build_typelib ((GIrNode *)type->parameter_type1, 
-					       module, modules, strings, types, 
-					       data, &pos, offset2);
+					       node, build, &pos, offset2);
 		      g_ir_node_build_typelib ((GIrNode *)type->parameter_type2, 
-					       module, modules, strings, types, 
-					       data, &pos, offset2);
+					       node, build, &pos, offset2);
 		    }
 		    break;
 		    
@@ -1506,7 +1564,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 		      else
 			blob->n_domains = 0;
 		      
-		      *offset2 = ALIGN_VALUE (*offset2 + 4 + 2 * blob->n_domains, 4);
+		      *offset2 = ALIGN_VALUE (*offset2 + G_STRUCT_OFFSET (ErrorTypeBlob, domains)
+		                              + 2 * blob->n_domains, 4);
 		      for (i = 0; i < blob->n_domains; i++)
 			blob->domains[i] = find_entry (module, modules, type->errors[i]);
 		    }
@@ -1527,7 +1586,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	FieldBlob *blob;
 
 	blob = (FieldBlob *)&data[*offset];
-	*offset += 8;
+	/* We handle the size member specially below, so subtract it */
+	*offset += sizeof (FieldBlob) - sizeof (SimpleTypeBlob);
 
 	blob->name = write_string (node->name, strings, data, offset2);
 	blob->readable = field->readable;
@@ -1540,8 +1600,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  blob->struct_offset = 0xFFFF; /* mark as unknown */
 
         g_ir_node_build_typelib ((GIrNode *)field->type, 
-				 module, modules, strings, types,
-				 data, offset, offset2);
+				 node, build, offset, offset2);
       }
       break;
 
@@ -1549,7 +1608,8 @@ g_ir_node_build_typelib (GIrNode    *node,
       {
 	GIrNodeProperty *prop = (GIrNodeProperty *)node;
 	PropertyBlob *blob = (PropertyBlob *)&data[*offset];
-	*offset += 8;
+        /* We handle the size member specially below, so subtract it */
+	*offset += sizeof (PropertyBlob) - sizeof (SimpleTypeBlob);
 
 	blob->name = write_string (node->name, strings, data, offset2);
 	blob->deprecated = prop->deprecated;
@@ -1560,8 +1620,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->reserved = 0;
 
         g_ir_node_build_typelib ((GIrNode *)prop->type, 
-				 module, modules, strings, types,
-				 data, offset, offset2);
+				 node, build, offset, offset2);
       }
       break;
 
@@ -1576,8 +1635,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (function->parameters);
 
-	*offset += sizeof(FunctionBlob);
-	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
+	*offset += sizeof (FunctionBlob);
+	*offset2 += sizeof (SignatureBlob) + n * sizeof (ArgBlob);
 
 	blob->blob_type = BLOB_TYPE_FUNCTION;
 	blob->deprecated = function->deprecated;
@@ -1595,8 +1654,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	g_debug ("building function '%s'", function->symbol);
 
         g_ir_node_build_typelib ((GIrNode *)function->result->type, 
-				 module, modules, strings, types,
-				 data, &signature, offset2);
+				 node, build, &signature, offset2);
 
 	blob2->may_return_null = function->result->allow_none;
 	blob2->caller_owns_return_value = function->result->transfer;
@@ -1610,9 +1668,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  {
 	    GIrNode *param = (GIrNode *)l->data;
 
-	    g_ir_node_build_typelib (param, 
-				     module, modules, strings, types,
-				     data, &signature, offset2);
+	    g_ir_node_build_typelib (param, node, build, &signature, offset2);
 	  }
 
       }
@@ -1629,8 +1685,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (function->parameters);
 
-	*offset += sizeof(CallbackBlob);
-	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
+	*offset += sizeof (CallbackBlob);
+	*offset2 += sizeof (SignatureBlob) + n * sizeof (ArgBlob);
 
 	blob->blob_type = BLOB_TYPE_CALLBACK;
 	blob->deprecated = function->deprecated;
@@ -1639,8 +1695,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->signature = signature;
 	
         g_ir_node_build_typelib ((GIrNode *)function->result->type, 
-				 module, modules, strings, types,
-				 data, &signature, offset2);
+				 node, build, &signature, offset2);
 
 	blob2->may_return_null = function->result->allow_none;
 	blob2->caller_owns_return_value = function->result->transfer;
@@ -1654,9 +1709,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  {
 	    GIrNode *param = (GIrNode *)l->data;
 
-	    g_ir_node_build_typelib (param, 
-				     module, modules, strings, types,
-				     data, &signature, offset2);
+	    g_ir_node_build_typelib (param, node, build, &signature, offset2);
 	  }
       }
       break;
@@ -1672,8 +1725,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (signal->parameters);
 
-	*offset += sizeof(SignalBlob);
-	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
+	*offset += sizeof (SignalBlob);
+	*offset2 += sizeof (SignatureBlob) + n * sizeof (ArgBlob);
 
 	blob->deprecated = signal->deprecated;
 	blob->run_first = signal->run_first;
@@ -1691,8 +1744,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->signature = signature;
 	
         g_ir_node_build_typelib ((GIrNode *)signal->result->type, 
-				 module, modules, strings, types,
-				 data, &signature, offset2);
+				 node, build, &signature, offset2);
 
 	blob2->may_return_null = signal->result->allow_none;
 	blob2->caller_owns_return_value = signal->result->transfer;
@@ -1706,8 +1758,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  {
 	    GIrNode *param = (GIrNode *)l->data;
 
-	    g_ir_node_build_typelib (param, module, modules, strings, types,
-				     data, &signature, offset2);
+	    g_ir_node_build_typelib (param, node, build, &signature, offset2);
 	  }
       }
       break;
@@ -1723,8 +1774,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (vfunc->parameters);
 
-	*offset += sizeof(VFuncBlob);
-	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
+	*offset += sizeof (VFuncBlob);
+	*offset2 += sizeof (SignatureBlob) + n * sizeof (ArgBlob);
 
 	blob->name = write_string (node->name, strings, data, offset2);
 	blob->must_chain_up = 0; /* FIXME */
@@ -1733,13 +1784,24 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->class_closure = 0; /* FIXME */
 	blob->reserved = 0;
 
+	if (vfunc->invoker)
+	  {
+	    int index = get_index_of_member_type ((GIrNodeInterface*)parent, G_IR_NODE_FUNCTION, vfunc->invoker);
+	    if (index == -1)
+	      {
+	        g_error ("Unknown member function %s for vfunc %s", vfunc->invoker, node->name);
+	      }
+            blob->invoker = (guint) index;
+	  }
+	else
+	  blob->invoker = 0x3ff; /* max of 10 bits */
+
 	blob->struct_offset = vfunc->offset;
 	blob->reserved2 = 0;
 	blob->signature = signature;
 	
         g_ir_node_build_typelib ((GIrNode *)vfunc->result->type, 
-				 module, modules, strings, types,
-				 data, &signature, offset2);
+				 node, build, &signature, offset2);
 
 	blob2->may_return_null = vfunc->result->allow_none;
 	blob2->caller_owns_return_value = vfunc->result->transfer;
@@ -1753,8 +1815,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  {
 	    GIrNode *param = (GIrNode *)l->data;
 
-	    g_ir_node_build_typelib (param, module, modules, strings, 
-				     types, data, &signature, offset2);
+	    g_ir_node_build_typelib (param, node, build, &signature, offset2);
 	  }
       }
       break;
@@ -1763,8 +1824,11 @@ g_ir_node_build_typelib (GIrNode    *node,
       {
 	ArgBlob *blob = (ArgBlob *)&data[*offset];
 	GIrNodeParam *param = (GIrNodeParam *)node;
-	
-	*offset += 12;
+
+	/* The offset for this one is smaller than the struct because
+	 * we recursively build the simple type inline here below.
+	 */
+	*offset += sizeof (ArgBlob) - sizeof (SimpleTypeBlob);
 
  	blob->name = write_string (node->name, strings, data, offset2);
  	blob->in = param->in;
@@ -1780,8 +1844,7 @@ g_ir_node_build_typelib (GIrNode    *node,
         blob->closure = param->closure;
         blob->destroy = param->destroy;
         
-        g_ir_node_build_typelib ((GIrNode *)param->type, module, modules, 
-				 strings, types, data, offset, offset2);
+        g_ir_node_build_typelib ((GIrNode *)param->type, node, build, offset, offset2);
       }
       break;
 
@@ -1793,6 +1856,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	
 	blob->blob_type = BLOB_TYPE_STRUCT;
 	blob->deprecated = struct_->deprecated;
+	blob->is_gtype_struct = struct_->is_gtype_struct;
 	blob->reserved = 0;
 	blob->name = write_string (node->name, strings, data, offset2);
 	blob->alignment = struct_->alignment;
@@ -1814,17 +1878,15 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->n_fields = 0;
 	blob->n_methods = 0;
 
-	*offset += 24; 
+	*offset += sizeof (StructBlob);
 
 	members = g_list_copy (struct_->members);
 
 	g_ir_node_build_members (&members, G_IR_NODE_FIELD, &blob->n_fields,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+	                         node, build, offset, offset2);
 
 	g_ir_node_build_members (&members, G_IR_NODE_FUNCTION, &blob->n_methods,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	g_ir_node_check_unhandled_members (&members, node->type);
 
@@ -1851,17 +1913,15 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->n_fields = 0;
 	blob->n_methods = 0;
 
-	*offset += 24; 
+	*offset += sizeof (StructBlob);
 
 	members = g_list_copy (boxed->members);
 
 	g_ir_node_build_members (&members, G_IR_NODE_FIELD, &blob->n_fields,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	g_ir_node_build_members (&members, G_IR_NODE_FUNCTION, &blob->n_methods,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	g_ir_node_check_unhandled_members (&members, node->type);
 
@@ -1899,30 +1959,29 @@ g_ir_node_build_typelib (GIrNode    *node,
 
 	blob->discriminator_offset = union_->discriminator_offset;
 
+	/* We don't support Union discriminators right now. */
+	/*
 	if (union_->discriminator_type)
 	  {
 	    *offset += 28;
 	    blob->discriminated = TRUE;
 	    g_ir_node_build_typelib ((GIrNode *)union_->discriminator_type, 
-				     module, modules, strings, types,
-				     data, offset, offset2);
+				     build, offset, offset2);
 	  }
-	else 
+	else
 	  {
-	    *offset += 32;
-	    blob->discriminated = FALSE;
-	    blob->discriminator_type.offset = 0;
-	  }
+        */
+	*offset += sizeof (UnionBlob);
+	blob->discriminated = FALSE;
+	blob->discriminator_type.offset = 0;
 	
 	members = g_list_copy (union_->members);
 
 	g_ir_node_build_members (&members, G_IR_NODE_FIELD, &blob->n_fields,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	g_ir_node_build_members (&members, G_IR_NODE_FUNCTION, &blob->n_functions,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	g_ir_node_check_unhandled_members (&members, node->type);
 
@@ -1934,8 +1993,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	      {
 		GIrNode *member = (GIrNode *)l->data;
 		
-		g_ir_node_build_typelib (member, module, modules, strings, 
-					 types, data, offset, offset2);
+		g_ir_node_build_typelib (member, node, build, offset, offset2);
 	      }
 	  }
       }
@@ -1947,8 +2005,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	EnumBlob *blob = (EnumBlob *)&data[*offset];
 	GIrNodeEnum *enum_ = (GIrNodeEnum *)node;
 
-	*offset += 20; 
-	
+	*offset += sizeof (EnumBlob); 
+
 	if (node->type == G_IR_NODE_ENUM)
 	  blob->blob_type = BLOB_TYPE_ENUM;
 	else
@@ -1979,8 +2037,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	    GIrNode *value = (GIrNode *)l->data;
 
 	    blob->n_values++;
-	    g_ir_node_build_typelib (value, module, modules, strings, types,
-				     data, offset, offset2);
+	    g_ir_node_build_typelib (value, node, build, offset, offset2);
 	  }
       }
       break;
@@ -2002,6 +2059,10 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  blob->parent = find_entry (module, modules, object->parent);
 	else
 	  blob->parent = 0;
+	if (object->glib_type_struct)
+	  blob->gtype_struct = find_entry (module, modules, object->glib_type_struct);
+	else
+	  blob->gtype_struct = 0;
 
 	blob->n_interfaces = 0;
 	blob->n_fields = 0;
@@ -2011,7 +2072,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->n_vfuncs = 0;
 	blob->n_constants = 0;
 	
-	*offset += 32;
+	*offset += sizeof(ObjectBlob);
 	for (l = object->interfaces; l; l = l->next)
 	  {
 	    blob->n_interfaces++;
@@ -2023,33 +2084,27 @@ g_ir_node_build_typelib (GIrNode    *node,
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_FIELD, &blob->n_fields,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_PROPERTY, &blob->n_properties,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_FUNCTION, &blob->n_methods,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_SIGNAL, &blob->n_signals,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_VFUNC, &blob->n_vfuncs,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_CONSTANT, &blob->n_constants,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	g_ir_node_check_unhandled_members (&members, node->type);
 
@@ -2069,6 +2124,10 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->name = write_string (node->name, strings, data, offset2);
 	blob->gtype_name = write_string (iface->gtype_name, strings, data, offset2);
 	blob->gtype_init = write_string (iface->gtype_init, strings, data, offset2);
+	if (iface->glib_type_struct)
+	  blob->gtype_struct = find_entry (module, modules, iface->glib_type_struct);
+	else
+	  blob->gtype_struct = 0;
 	blob->n_prerequisites = 0;
 	blob->n_properties = 0;
 	blob->n_methods = 0;
@@ -2076,7 +2135,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->n_vfuncs = 0;
 	blob->n_constants = 0;
 	
-	*offset += 28;
+	*offset += sizeof (InterfaceBlob);
 	for (l = iface->prerequisites; l; l = l->next)
 	  {
 	    blob->n_prerequisites++;
@@ -2088,28 +2147,23 @@ g_ir_node_build_typelib (GIrNode    *node,
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_PROPERTY, &blob->n_properties,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_FUNCTION, &blob->n_methods,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_SIGNAL, &blob->n_signals,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_VFUNC, &blob->n_vfuncs,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	*offset = ALIGN_VALUE (*offset, 4);
 	g_ir_node_build_members (&members, G_IR_NODE_CONSTANT, &blob->n_constants,
-				 module, modules, strings,
-				 types, data, offset, offset2);
+				 node, build, offset, offset2);
 
 	g_ir_node_check_unhandled_members (&members, node->type);
 
@@ -2122,7 +2176,7 @@ g_ir_node_build_typelib (GIrNode    *node,
       {
 	GIrNodeValue *value = (GIrNodeValue *)node;
 	ValueBlob *blob = (ValueBlob *)&data[*offset];
-	*offset += 12;
+	*offset += sizeof (ValueBlob);
 
 	blob->deprecated = value->deprecated;
 	blob->reserved = 0;
@@ -2135,7 +2189,7 @@ g_ir_node_build_typelib (GIrNode    *node,
       {
 	GIrNodeErrorDomain *domain = (GIrNodeErrorDomain *)node;
 	ErrorDomainBlob *blob = (ErrorDomainBlob *)&data[*offset];
-	*offset += 16;
+	*offset += sizeof (ErrorDomainBlob);
 
 	blob->blob_type = BLOB_TYPE_ERROR_DOMAIN;
 	blob->deprecated = domain->deprecated;
@@ -2153,8 +2207,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	ConstantBlob *blob = (ConstantBlob *)&data[*offset];
 	guint32 pos;
 
-	pos = *offset + 8;
-	*offset += 20;
+	pos = *offset + G_STRUCT_OFFSET (ConstantBlob, type);
+	*offset += sizeof (ConstantBlob);
 
 	blob->blob_type = BLOB_TYPE_CONSTANT;
 	blob->deprecated = constant->deprecated;
@@ -2200,6 +2254,14 @@ g_ir_node_build_typelib (GIrNode    *node,
 	    blob->size = 8;
 	    *(guint64*)&data[blob->offset] = (guint64) parse_uint_value (constant->value);
 	    break;
+	  case GI_TYPE_TAG_SHORT:
+	    blob->size = sizeof (gshort);
+	    *(gshort*)&data[blob->offset] = (gshort) parse_int_value (constant->value);
+	    break;
+	  case GI_TYPE_TAG_USHORT:
+	    blob->size = sizeof (gushort);
+	    *(gushort*)&data[blob->offset] = (gushort) parse_uint_value (constant->value);
+	    break;
 	  case GI_TYPE_TAG_INT:
 	    blob->size = sizeof (gint);
 	    *(gint*)&data[blob->offset] = (gint) parse_int_value (constant->value);
@@ -2235,8 +2297,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  }
 	*offset2 += ALIGN_VALUE (blob->size, 4);
 	
-	g_ir_node_build_typelib ((GIrNode *)constant->type, module, modules, 
-				 strings, types, data, &pos, offset2);
+	g_ir_node_build_typelib ((GIrNode *)constant->type, node, build, &pos, offset2);
       }
       break;
     default:
@@ -2251,7 +2312,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	   old_offset, *offset, old_offset2, *offset2);
 
   if (*offset2 - old_offset2 + *offset - old_offset > g_ir_node_get_full_size (node))
-    g_error ("exceeding space reservation !!");
+    g_error ("exceeding space reservation; offset: %d (prev %d) offset2: %d (prev %d) nodesize: %d",
+             *offset, old_offset, *offset2, old_offset2, g_ir_node_get_full_size (node));
 }
 
 /* if str is already in the pool, return previous location, otherwise write str
