@@ -1,6 +1,6 @@
 #include "annotation.h"
 
-static char backslash_parsing_tester = '\\';
+char backslash_parsing_tester = '\\';
 
 G_DEFINE_TYPE (AnnotationObject, annotation_object, G_TYPE_OBJECT);
 
@@ -408,7 +408,7 @@ gboolean
 annotation_object_string_out(AnnotationObject *object,
                              char            **str_out)
 {
-
+  return FALSE;
 }
 
 /**
@@ -512,6 +512,41 @@ annotation_object_do_not_use (AnnotationObject *object)
   return NULL;
 }
 
+/**
+ * annotation_object_watch:
+ * @object: A #AnnotationObject
+ * @func: The callback
+ * @user_data: The callback data
+ *
+ * This is here just for the sake of being overriden by its
+ * annotation_object_watch_full().
+ */
+void
+annotation_object_watch (AnnotationObject *object,
+                         AnnotationForeachFunc func,
+                         gpointer user_data)
+{
+}
+
+/**
+ * annotation_object_watch_full:
+ * @object: A #AnnotationObject
+ * @func: The callback
+ * @user_data: The callback data
+ * @destroy: Destroy notification
+ *
+ * Test overriding via the "Rename To" annotation.
+ *
+ * Rename to: annotation_object_watch
+ */
+void
+annotation_object_watch_full (AnnotationObject *object,
+                              AnnotationForeachFunc func,
+                              gpointer user_data,
+                              GDestroyNotify destroy)
+{
+}
+
 /**
  * annotation_init:
  * @argc: (inout): The number of args. 
@@ -553,7 +588,6 @@ annotation_string_zero_terminated (void)
 void
 annotation_string_zero_terminated_out (char ***out)
 {
-  return NULL;
 }
 
 /**
@@ -566,5 +600,50 @@ annotation_versioned (void)
 {
 }
 
+/**
+ * annotation_object_extra_annos:
+ *
+ * Attributes: (org.foobar testvalue)
+ */
+void
+annotation_object_extra_annos (AnnotationObject *object)
+{
+}
 
-static char backslash_parsing_tester_2 = '\\';
+/**
+ * annotation_custom_destroy:
+ *
+ * @callback: (destroy destroy) (closure data): Destroy notification
+ *
+ * Test messing up the heuristic of closure/destroy-notification
+ * detection, and fixing it via annotations.
+ */
+void
+annotation_custom_destroy (AnnotationCallback callback,
+                           AnnotationNotifyFunc destroy,
+                           gpointer data)
+{
+}
+
+/**
+ * annotation_get_source_file:
+ *
+ * Return value: (type filename): Source file
+ */
+char *
+annotation_get_source_file (void)
+{
+  return NULL;
+}
+
+/**
+ * annotation_set_source_file:
+ * @fname: (type filename): Source file
+ *
+ */
+void
+annotation_set_source_file (const char *fname)
+{
+}
+
+char backslash_parsing_tester_2 = '\\';
